import numpy as np
import rasterio as rio
from odc.geo.geobox import GeoBox

from raster_tools.masking import get_default_null_value
from raster_tools.raster import Raster, get_raster
from raster_tools.utils import make_raster_ds

__all__ = [
    "reproject",
]

SUPPORTED_RESAMPLE_METHODS = {m.name for m in rio.warp.SUPPORTED_RESAMPLING}


def reproject(
    raster, crs_or_geobox=None, resample_method="nearest", resolution=None
):
    """Reproject to a new projection or resolution.

    This is a lazy operation.

    Parameters
    ----------
    raster : str, Raster
        The raster to reproject.
    crs_or_geobox : int, str, CRS, GeoBox, optional
        The target grid to reproject the raster to. This can be a projection
        string, EPSG code string or integer, a CRS object, or a GeoBox object.
        `resolution` can also be specified to change the output raster's
        resolution in the new CRS. If `crs_or_geobox` is not provided,
        `resolution` must be specified.
    resample_method : str, optional
        The data resampling method to use. Null pixels are ignored for all
        methods. Some methods require specific versions of GDAL. These are
        noted below. Valid methods are:

        'nearest'
            Nearest neighbor resampling. This is the default.
        'bilinear'
            Bilinear resmapling.
        'cubic'
            Cubic resmapling.
        'cubic_spline'
            Cubic spline resmapling.
        'lanczos'
            Lanczos windowed sinc resmapling.
        'average'
            Average resampling, computes the weighted average of all
            contributing pixels.
        'mode'
            Mode resampling, selects the value which appears most often.
        'max'
            Maximum resampling. (GDAL 2.0+)
        'min'
            Minimum resampling. (GDAL 2.0+)
        'med'
            Median resampling. (GDAL 2.0+)
        'q1'
            Q1, first quartile resampling. (GDAL 2.0+)
        'q3'
            Q3, third quartile resampling. (GDAL 2.0+)
        'sum'
            Sum, compute the weighted sum. (GDAL 3.1+)
        'rms'
            RMS, root mean square/quadratic mean. (GDAL 3.3+)
    resolution : int, float, tuple of int or float, optional
        The desired resolution of the reprojected raster. If `crs_or_geobox` is
        unspecified, this is used to reproject to the new resolution while
        maintaining the same CRS. One of `crs_or_geobox` or `resolution` must
        be provided. Both can also be provided.

    Returns
    -------
    Raster
        The reprojected raster on the new grid.

    """
    raster = get_raster(raster)
    if resample_method not in SUPPORTED_RESAMPLE_METHODS:
        raise ValueError(
            f"Unsupported resampling method provided: {resample_method!r}. "
            "Supported methods: {}".format(
                ", ".join(sorted(SUPPORTED_RESAMPLE_METHODS))
            )
        )
    if crs_or_geobox is None:
        if resolution is None:
            raise ValueError("Must supply either crs_or_geobox or resolution")
        dst_gb = raster.geobox
    elif not isinstance(crs_or_geobox, GeoBox):
        dst_gb = raster.geobox.to_crs(crs_or_geobox)
    else:
        dst_gb = crs_or_geobox
    if resolution is not None:
        if resolution <= 0:
            raise ValueError("Resolution must be a postive value")
        dst_gb = dst_gb.zoom_to(resolution=resolution)
    if dst_gb == raster.geobox:
        return raster.copy()

    nv = (
        raster.null_value
        if raster._masked
        else get_default_null_value(raster.dtype)
    )
    reprojected = raster._ds.raster.odc.reproject(
        dst_gb, resampling=resample_method, dst_nodata=nv
    ).rio.write_nodata(nv)
    if "longitude" in reprojected.dims:
        # odc-geo will rename x/y to lon/lat for lon/lat based projections, so
        # revert to x/y
        reprojected = reprojected.rename({"longitude": "x", "latitude": "y"})
    mask = reprojected == nv if not np.isnan(nv) else np.isnan(reprojected)
    ds = make_raster_ds(reprojected, mask)
    return Raster(ds, _fast_path=True)



from collections.abc import Sequence
from functools import partial

import dask
import dask.array as da
import dask.dataframe as dd
import dask_geopandas as dgpd
import geopandas as gpd
import numba as nb
import numpy as np
import pandas as pd

from raster_tools.dtypes import F64, I64, is_int, is_str
from raster_tools.raster import Raster, get_raster
from raster_tools.vector import Vector, get_vector

__all__ = ["ZONAL_STAT_FUNCS", "extract_points_eager", "zonal_stats"]


@nb.jit(nopython=True, nogil=True)
def _entropy(counts):
    if len(counts) == 0:
        return np.nan
    # Manual stat calculation with a loop is faster than using vectorized numpy
    # operations, likely due to numpy ops allocating temporary arrays.
    entropy = 0.0
    frac = 1 / np.sum(counts)
    for cnt in counts:
        p = cnt * frac
        entropy -= p * np.log(p)
    return entropy


@nb.jit(nopython=True, nogil=True)
def _asm(counts):
    if len(counts) == 0:
        return np.nan
    # Manual stat calculation with a loop is faster than using vectorized numpy
    # operations, likely due to numpy ops allocating temporary arrays.
    asm = 0.0
    frac = 1 / np.sum(counts)
    for cnt in counts:
        p = cnt * frac
        asm += p * p
    return asm


def _unique_with_counts_chunk(grps):
    return grps.apply(lambda s: np.unique(s, return_counts=True))


def _combine_values_and_counts(s):
    result = (
        pd.DataFrame(
            {
                "values_": [vs for vs, _ in s.to_numpy()],
                "counts_": [cs for _, cs in s.to_numpy()],
            }
        )
        .explode(["values_", "counts_"])
        .groupby("values_")
        .sum()
    )
    # Convert to lists because of weird behavior in finalize step. If the
    # results are not converted to lists, the finalize step will get a series
    # where each row is a tuple of arrays of objects instead of a tuple of
    # lists of numbers. I don't know why this happens, but the following works.
    return (result.index.to_list(), result.counts_.to_list())


def _mode_fin(s):
    # Final operation of the mode calculation
    results = np.empty(len(s), dtype=F64)
    for i, (vs, cs) in enumerate(s.to_numpy()):
        if len(vs) != 0:
            # Order the counts descending with values ascending. In the case of
            # tied counts, lower values come first. This mirrors the behavior
            # of scipy's mode function.
            values_and_counts = pd.DataFrame(
                {"values_": vs, "counts_": cs}
            ).sort_values(by=["counts_", "values_"], ascending=[False, True])
            results[i] = values_and_counts.values_.iloc[0]
        else:
            results[i] = np.nan
    return pd.Series(results, index=s.index.copy())


def _asm_entropy_fin(s, func):
    # Final operation of the asm and entropy calculations
    if len(s) == 0:
        return pd.Series(np.array((), dtype=F64), index=s.index)
    result_list = []
    for _, cs in s.to_numpy():
        result_list.append(func(np.asarray(cs)))
    return pd.Series(result_list, index=s.index.copy())


# Notes for mode, asm, and entropy aggs:
# chunk:
#     Each row is now a 2-tuple of numpy arrays representing the unique values
#     and their counts
# agg:
#     Compbine the values and counts for zone value into a single 2-tuple of
#     values and counts lists.
# finalize:
#     Compute final stat from values and counts.
_mode_agg = dd.Aggregation(
    "mode",
    chunk=lambda grps: grps.apply(lambda s: np.unique(s, return_counts=True)),
    agg=lambda grps: grps.apply(_combine_values_and_counts),
    finalize=_mode_fin,
)
_asm_agg = dd.Aggregation(
    "asm",
    chunk=lambda grps: grps.apply(lambda s: np.unique(s, return_counts=True)),
    agg=lambda grps: grps.apply(_combine_values_and_counts),
    finalize=partial(_asm_entropy_fin, func=_asm),
)
_entropy_agg = dd.Aggregation(
    "entropy",
    chunk=_unique_with_counts_chunk,
    agg=lambda grps: grps.apply(_combine_values_and_counts),
    finalize=partial(_asm_entropy_fin, func=_entropy),
)
_nunique_agg = dd.Aggregation(
    name="nunique",
    chunk=lambda s: s.apply(lambda x: list(set(x.dropna()))),
    agg=lambda s0: s0.obj.groupby(
        level=list(range(s0.obj.index.nlevels))
    ).sum(),
    finalize=lambda s1: s1.apply(lambda final: len(set(final))),
)


_DASK_STAT_NAMES = frozenset(
    (
        "count",
        "max",
        "mean",
        "median",
        "min",
        "prod",
        "size",
        "std",
        "sum",
        "var",
    )
)
_CUSTOM_STAT_NAMES = frozenset(("asm", "entropy", "mode", "nunique"))
_CUSTOM_STAT_AGGS = {
    "asm": _asm_agg,
    "entropy": _entropy_agg,
    "mode": _mode_agg,
    "nunique": _nunique_agg,
}
ZONAL_STAT_FUNCS = frozenset(sorted(_DASK_STAT_NAMES | _CUSTOM_STAT_NAMES))


def _build_long_format_meta(df):
    return pd.DataFrame(
        {
            "zone": np.array((), dtype=df.index.dtype),
            "band": np.array((), dtype=I64),
            **{
                s: np.array((), dtype=dtype)
                for s, dtype in df.dtypes["band_1"].items()
            },
        },
        index=np.array((), dtype=I64),
    )


def _melt_part(part):
    # Convert a dataframe from wide format to long format
    #
    # DataFrame structure, the columns are a MultiIndex and zone ids
    # are the index:
    #       band_1             band_2             ...
    #       stat1  stat2  ...  stat1  stat2  ...  ...
    # zone
    #    1     --     --  ...
    part = (
        # Unpivot band_# column labels into the index. The index is now a
        # MultiIndex of (zone, level_1) where level_1 has the band labels
        # The PD013 warning is silenced below because I could not get melt to
        # reproduce the desired result (I tried A LOT).
        part.stack(0, future_stack=True)  # noqa: PD013
        # Move the zone and level_1 indices to columns
        .reset_index()
        .rename(columns={"level_1": "band"})
        .sort_values(["band", "zone"])
    )
    # Convert band labels to 1-based integers
    part["band"] = part.band.apply(lambda x: int(x.split("_")[-1]))
    # New DataFrame structure:
    #        zone  band  stat1  stat2  ...
    # index
    #     0     1     1     --     --  ...
    return part


def _find_problem_stats(stats):
    has_median = False
    out = []
    for s in stats:
        if s != "median":
            out.append(s)
        else:
            has_median = True
    return out, has_median


def _raster_to_series(raster):
    # Convert the underlying dask array into a dask Series
    # DataArray.to_dask_dataframe is only available in newer versions of xarray
    if hasattr(raster.xdata, "to_dask_dataframe"):
        return raster.xdata.to_dask_dataframe(["band", "x", "y"])["raster"]
    return raster._ds.to_dask_dataframe(["band", "x", "y"])["raster"]


def _zonal_stats(features_raster, data_raster, stats):
    """
    Convert data to large dask DataFrame, group by the feature zone ids, and
    apply aggregations specified by `stats` for each band.
    """
    stat_names = stats
    stats, has_median = _find_problem_stats(stats)
    stats = [
        s if s in _DASK_STAT_NAMES else _CUSTOM_STAT_AGGS[s] for s in stats
    ]
    if features_raster.data.chunks[1:] != data_raster.data.chunks[1:]:
        # Force rasters to have matching chunksizes
        features_raster = features_raster.chunk(data_raster.data.chunks)
    # Convert the features raster and data raster bands to dataframes and join
    # them together. Each cell becomes a row and each chunk becomes a
    # partition. The order is the same and the chunk boundaries are the same so
    # they should concat together, cleanly.
    raster_dfs = [_raster_to_series(features_raster).rename("zone")]
    for b in range(1, data_raster.nbands + 1):
        band = _raster_to_series(data_raster.get_bands(b)).rename(f"band_{b}")
        if data_raster.null_value is not None and not np.isnan(
            data_raster.null_value
        ):
            # Replace null values with NA values
            band = band.replace(data_raster.null_value, np.nan)
        # Cast up to avoid floating point issues in sums. F32 sums loose
        # precision quickly even for smaller rasters. This affects any stat
        # that involves a sum such as std, var, and mean.
        # ref: https://web.archive.org/web/20230329091023/https://pythonspeed.com/articles/float64-float32-precision/  # noqa
        if band.dtype != F64:
            band = band.astype(F64)
        raster_dfs.append(band)
    # DataFrame structure:
    #        zone  band_1  band_2  ...
    # index
    #     0    --      --      --  ...
    combined_raster_df = dd.concat(raster_dfs, axis=1)
    # Filter out non-feature areas
    combined_raster_df = combined_raster_df[
        combined_raster_df.zone != features_raster.null_value
    ]
    grouped = combined_raster_df.groupby("zone")
    agg_result_df = None
    if has_median:
        # median requires special treatment. It needs the shuffle arg and also
        # causes TypeErrors when used with custom agg functions.
        # ref: https://github.com/dask/dask/issues/10517
        median_df = grouped.agg(["median"], shuffle_method="tasks")
    if len(stats):
        agg_result_df = grouped.agg(stats)
        if has_median:
            agg_result_df = agg_result_df.join(median_df)
    else:
        # Only median was provided
        agg_result_df = median_df
    if has_median:
        # Shuffle columns to be in order of the original stats
        col_tuples = []
        for b in range(1, data_raster.nbands + 1):
            band = f"band_{b}"
            col_tuples.extend((band, stat) for stat in stat_names)
        target_columns = pd.MultiIndex.from_tuples(col_tuples)
        if not agg_result_df.columns.equals(target_columns):
            agg_result_df = agg_result_df[target_columns]
    # DataFrame structure, the columns are a MultiIndex and zone ids are the
    # index:
    #       band_1             band_2             ...
    #       stat1  stat2  ...  stat1  stat2  ...  ...
    # zone
    #    1     --     --  ...
    return agg_result_df


def zonal_stats(
    features, data_raster, stats, features_field=None, wide_format=True
):
    """Apply stat functions to a raster based on a set of features.

    Parameters
    ----------
    features : str, Vector, Raster
        A `Vector` or path string pointing to a vector file or a categorical
        Raster. The vector features are used like cookie cutters to pull data
        from the `data_raster` bands. If `features` is a Raster, it must be an
        int dtype and have only one band.
    data_raster : Raster, str
        A `Raster` or path string pointing to a raster file. The data raster
        to pull data from and apply the stat functions to.
    stats : str, list of str
        A single string or list of strings corresponding to stat functions.
        These functions will be applied to the raster data for each of the
        features in `features`. Valid string values:

        'asm'
            Angular second moment. Applies -sum(P(g)**2) where P(g) gives the
            probability of g within the neighborhood.
        'count'
            Count valid cells.
        'entropy'
            Calculates the entropy. Applies -sum(P(g) * log(P(g))). See 'asm'
            above.
        'max'
            Find the maximum value.
        'mean'
            Calculate the mean.
        'median'
            Calculate the median value.
        'min'
            Find the minimum value.
        'mode'
            Compute the statistical mode of the data. In the case of a tie, the
            lowest value is returned.
        'nunique'
            Count unique values.
        'prod'
            Calculate the product.
        'size'
            Calculate zone size.
        'std'
            Calculate the standard deviation.
        'sum'
            Calculate the sum.
        'var'
            Calculate the variance.
    features_field : str, optional
        If the `features` argument is a vector, this determines which field to
        use when rasterizing the features. It must match one of the fields in
        `features`. The default is to use `features`' index.
    wide_format : bool, optional
        If ``True``, the resulting dataframe is returned in wide format where
        the columns are a cartesian product of the `data_raster` bands and the
        specified stats and the index contains the feature zone IDs.

        .. code-block::

            pandas.MultiIndex(
              [
                ('band_1', 'stat1'),
                ('band_1', 'stat2'),
                ...
                ('band_2', 'stat1'),
                ('band_2', 'stat2'),
                ...
              ],
            )

        If ``False``, the resulting dataframe has columns `'zone', 'band',
        'stat1', 'stat2', ...` and an integer index. In this case, the zone
        column contains the feature zone IDs and band contains the one-base
        integer band number. The rest of the columns correspond to the
        specified stats.

        The default is wide format.

    Returns
    -------
    dask.dataframe.DataFrame
        A delayed dask DataFrame where the specified stats have been applied to
        the bands in `data_raster`. See the `wide_format` option for a
        description of the dataframe's structure.

    """
    if isinstance(
        features,
        (
            str,
            Vector,
            dgpd.GeoDataFrame,
            dgpd.GeoSeries,
            gpd.GeoDataFrame,
            gpd.GeoSeries,
        ),
    ):
        features = get_vector(features)
    elif isinstance(features, Raster):
        if not is_int(features.dtype):
            raise TypeError("Feature raster must be an integer type.")
        if features.shape[0] > 1:
            raise ValueError("Feature raster must have only 1 band.")
    else:
        raise TypeError(
            "Could not understand features arg. Must be Vector, str or Raster"
        )
    data_raster = get_raster(data_raster)
    if is_str(stats):
        stats = [stats]
    elif isinstance(stats, Sequence):
        stats = list(stats)
        if not stats:
            raise ValueError("No stat functions provide")
    else:
        raise ValueError(f"Could not understand stats arg: {repr(stats)}")
    for stat in stats:
        if stat not in ZONAL_STAT_FUNCS:
            raise ValueError(f"Invalid stats function: {repr(stat)}")
    if isinstance(features, Raster):
        if features.crs != data_raster.crs:
            raise ValueError("Feature raster CRS must match data raster")
        if features.shape[1:] != data_raster.shape[1:]:
            raise ValueError("Feature raster shape must match data raster")

    # Rechunk based on largest (probable) dtype to avoid overly large chunks,
    # which could cause memory issues down the pipeline. For instance, if the
    # data raster has dtype of f32 but the rasterization of the features
    # produces an i64 raster, the features raster will have double the memory
    # footprint, for each chunk, compared to the original data raster. This
    # causes dask to raise warnings about chunk sizes and drastically increases
    # the likelihood of running out of memory at compute time.
    # Rechunking to an 8-byte dtype helps mitigate the potential for memory
    # pressure at compute time. It is done here because the chunksize of the
    # data raster determines the chunksize of the features raster.
    new_chunksize = da.empty((1, *data_raster.shape[1:]), dtype=F64).chunksize
    data_raster = data_raster.chunk(new_chunksize)
    features_raster = None
    if isinstance(features, Vector):
        if features_field is not None and features_field not in features.data:
            raise KeyError(
                "features_field must be a field name in the features input"
            )
        features_raster = features.to_raster(data_raster, field=features_field)
    else:
        if features.nbands > 1:
            raise ValueError("features raster must have a single band")
        if features.shape[1:] != data_raster.shape[1:]:
            raise ValueError(
                "features raster shape must match the data raster. "
                f"Expected {data_raster.shape[1:]}, got {features.shape[1:]}."
            )
        features_raster = features

    zonal_result_df = _zonal_stats(features_raster, data_raster, stats)
    if not wide_format:
        # New DataFrame structure:
        #        zone  band  stat1  stat2  ...
        # index
        #     0     1     1     --     --  ...
        zonal_result_df = zonal_result_df.map_partitions(
            _melt_part, meta=_build_long_format_meta(zonal_result_df)
        )
    return zonal_result_df


def _create_dask_range_index(start, stop):
    # dask.dataframe only allows dask.dataframe.index objects but doesn't have
    # a way to create them. this is a hack to create one using from_pandas.
    dummy = pd.DataFrame(
        {"tmp": np.zeros(stop - start, dtype="u1")},
        index=pd.RangeIndex(start, stop),
    )
    return dd.from_pandas(dummy, 1).index


def extract_points_eager(
    points, raster, column_name="extracted", skip_validation=True, axis=0
):
    """Extract the raster cell values using point features

    Note
    ----
    This function is partially eager. The x and y values for the target points
    are computed. The result is still a lazy dask DataFrame.


    This finds the grid cells that the points fall into and extracts the value
    at each point. The input feature will be partially computed to make sure
    that all of the geometries are points, unless `skip_validation` is set to
    `True`.

    Parameters
    ----------
    points : str, Vector
        The points to use for extracting data.
    raster : str, Raster
        The raster to pull data from.
    column_name : str, optional
        The column name to use for the extracted data points. Default is
        `"extracted"`.
    skip_validation : bool, optional
        If `True`, the input `points` is not validated to make sure that all
        features are points. This prevents partially computing the data.
        Default is `True`.
    axis : int, optional
        If 0 band column and values will be appended to a dataframe. Otherwise
        band values will be append to the columns named after the prefix and
        band of a dataframe

    Returns
    -------
    dask.dataframe.DataFrame
        The columns names depend on the value of axis and are based on the
        "band" and `column_name variable. If axis = 0, the output band column
        within the dataframe identifies the band the value was extracted from.
        The values within the column named after the column name variable are
        the extracted values from the given band. Otherwise, the column names
        within the dataframe are appended to the column_name prefix and provide
        the extracted values. NaN values in the extracted column are where
        there was missing data in the raster or the point was outside the
        raster's domain.
    """
    points = get_vector(points)
    raster = get_raster(raster)

    if not len(column_name):
        raise ValueError("column_name must not be empty")
    if (
        not skip_validation
        and not (points.geometry.geom_type == "Point").all().compute()
    ):
        raise TypeError("All geometries must be points.")

    if raster.crs is not None and raster.crs != points.crs:
        gdf = points.to_crs(raster.crs).data
    else:
        gdf = points.data
    x = gdf.geometry.x.to_dask_array()
    y = gdf.geometry.y.to_dask_array()
    r, c = raster.index(*dask.compute(x, y))
    nb, nr, nc = raster.shape
    valid = (r >= 0) & (r < nr) & (c >= 0) & (c < nc)
    n = len(valid)
    dfs = []
    for i in range(nb):
        bnd = i + 1
        extracted = da.full(n, np.nan, dtype=F64)
        extracted[valid] = raster.data.vindex[i, r[valid], c[valid]]
        # Mask out missing points within the valid zones
        exmask = da.zeros(n, dtype=bool)
        exmask[valid] = raster.mask.vindex[i, r[valid], c[valid]]
        extracted[exmask] = np.nan
        if axis == 0:
            index = _create_dask_range_index(n * i, n * bnd)
            extracted_df = (
                da.full(n, bnd, dtype=np.min_scalar_type(nb + 1))
                .to_dask_dataframe(index=index)
                .to_frame("band")
            )
            extracted_df[column_name] = extracted.to_dask_dataframe(
                index=index
            )
        else:
            extracted_df = extracted.to_dask_dataframe(
                columns=column_name + "_" + str(bnd)
            )
        assert extracted_df.known_divisions
        dfs.append(extracted_df)
    extracted_result = dd.concat(dfs, axis=axis)
    return extracted_result



# isort: off
# TODO(pygeos): remove this once shapely is the default backend for geopandas.
# Force raster_tools._compat to be loaded before geopandas when running tests
import raster_tools as rts  # noqa: F401

# isort: on


import unittest

import dask
import dask_geopandas as dgpd
import geopandas as gpd
import numpy as np
import pytest
import rasterio as rio

from raster_tools.vector import _OBJECTID_BASE, Vector, open_vectors


class TestOpenVectors(unittest.TestCase):
    def test_open_vectors(self):
        vs = open_vectors("tests/data/vector/Zones.gdb")

        self.assertIsInstance(vs, list)
        self.assertTrue(len(vs) == 2)
        self.assertIsInstance(vs[0], Vector)
        self.assertIsInstance(vs[1], Vector)
        self.assertTrue(len(vs[0]) == 10642)
        self.assertTrue(len(vs[1]) == 184)

        pods1 = open_vectors("tests/data/vector/Zones.gdb", layers="PODs")
        self.assertIsInstance(pods1, Vector)
        pods2 = open_vectors("tests/data/vector/Zones.gdb", layers=1)
        pods3 = open_vectors("tests/data/vector/Zones.gdb", layers=["PODs"])
        pods4 = open_vectors("tests/data/vector/Zones.gdb", layers=[1])
        for v in [pods2, pods3, pods4]:
            self.assertTrue(pods1.data.compute().equals(v.data.compute()))

    def test_open_vectors_errors(self):
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers="dummy")
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers=2)
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers=-1)
        with self.assertRaises(ValueError):
            open_vectors("tests/data/vector/Zones.gdb", layers=[-1])

        with self.assertRaises(TypeError):
            open_vectors("tests/data/vector/Zones.gdb", layers=[0, "PODs"])
        with self.assertRaises(TypeError):
            open_vectors("tests/data/vector/Zones.gdb", layers={})


class TestVectorProperties(unittest.TestCase):
    def setUp(self):
        self.v = open_vectors("tests/data/vector/pods.shp")

    def test_table(self):
        self.assertTrue(hasattr(self.v, "table"))
        self.assertTrue(len(self.v) == len(self.v.table))
        # Table doesn't contain geometry. It is only vector attributes
        self.assertTrue("geometry" not in self.v.table.columns)
        self.assertTrue(
            self.v.table.columns.size == (self.v.data.columns.size - 1)
        )

    def test_size(self):
        self.assertTrue(hasattr(self.v, "size"))
        self.assertTrue(self.v.size == 184)

    def test_shape(self):
        self.assertTrue(hasattr(self.v, "shape"))
        self.assertIsInstance(self.v.shape, tuple)
        self.assertTrue(self.v.shape == dask.compute(self.v.table.shape)[0])

    def test_crs(self):
        self.assertTrue(hasattr(self.v, "crs"))
        self.assertIsInstance(self.v.crs, rio.crs.CRS)
        self.assertTrue(self.v.crs == rio.crs.CRS.from_epsg(5070))

    def test_field_schema(self):
        self.assertTrue(hasattr(self.v, "field_schema"))
        self.assertIsInstance(self.v.field_schema, dict)
        self.assertTrue(self.v.field_schema == self.v.table.dtypes.to_dict())

    def test_field_names(self):
        self.assertTrue(hasattr(self.v, "field_names"))
        self.assertIsInstance(self.v.field_names, list)
        self.assertTrue(self.v.field_names == self.v.table.columns.to_list())

    def test_field_dtypes(self):
        self.assertTrue(hasattr(self.v, "field_dtypes"))
        self.assertIsInstance(self.v.field_dtypes, list)
        self.assertTrue(
            self.v.field_dtypes == list(self.v.table.dtypes.to_dict().values())
        )

    def test_geometry(self):
        self.assertTrue(hasattr(self.v, "geometry"))
        self.assertIsInstance(self.v.geometry, dgpd.GeoSeries)
        self.assertTrue(
            (self.v.geometry == self.v.data.geometry).all().compute()
        )

    def test_tasks(self):
        assert hasattr(self.v, "tasks")
        v = self.v.copy()
        v._geo = v._geo.compute()
        assert not dask.is_dask_collection(v.data)
        assert v.tasks == 0
        v = self.v.to_lazy()
        assert v.tasks >= 1

    def test_bounds(self):
        self.assertTrue(hasattr(self.v, "bounds"))
        geoms_df = gpd.read_file("tests/data/vector/pods.shp")
        vect = open_vectors("tests/data/vector/pods.shp")
        self.assertTrue(all(vect.bounds == geoms_df.total_bounds))
        self.assertTrue(dask.is_dask_collection(vect.to_lazy().bounds))
        self.assertTrue(all(vect.to_lazy().bounds == geoms_df.total_bounds))


class TestSpecialMethods(unittest.TestCase):
    def setUp(self):
        self.v = open_vectors("tests/data/vector/pods.shp")

    def test_len(self):
        self.assertTrue(hasattr(self.v, "__len__"))
        self.assertTrue(len(self.v) == self.v.size)
        self.assertTrue(len(self.v) == len(self.v.data))

    def test_getitem(self):
        self.assertTrue(hasattr(self.v, "__getitem__"))
        self.assertIsInstance(self.v[0], Vector)
        self.assertTrue(
            self.v[0].data.compute().equals(self.v.data.loc[[0]].compute())
        )
        last = self.v.data.loc[[self.v.size - 1]]
        last.index = dask.array.from_array([0]).to_dask_dataframe()
        self.assertTrue(self.v[-1].data.compute().equals(last.compute()))
        with self.assertRaises(NotImplementedError):
            self.v[0:3]
        with self.assertRaises(TypeError):
            self.v[9.0]
        with self.assertRaises(IndexError):
            self.v[self.v.size]
        with self.assertRaises(IndexError):
            self.v[-200]
        with self.assertRaises(IndexError):
            self.v[999]


class TestCopy(unittest.TestCase):
    def test_copy(self):
        v = open_vectors("tests/data/vector/pods.shp")
        vc = v.copy()
        self.assertIsNot(v, vc)
        self.assertIsNot(v.data, vc.data)
        self.assertTrue(v.data.compute().equals(vc.data.compute()))


class TestLazyByDefault(unittest.TestCase):
    def test_lazy_by_default(self):
        v = open_vectors("tests/data/vector/pods.shp")
        self.assertTrue(dask.is_dask_collection(v._geo))


class TestEval(unittest.TestCase):
    def test_eval(self):
        v = open_vectors("tests/data/vector/Zones.gdb", 0)
        self.assertTrue(dask.is_dask_collection(v.data))
        ve = v.eval()
        self.assertTrue(dask.is_dask_collection(ve.data))
        self.assertTrue(ve.data.npartitions == 1)


class TestConversions(unittest.TestCase):
    def setUp(self):
        self.v = open_vectors("tests/data/vector/pods.shp")

    def test_to_lazy(self):
        vc = self.v.copy()
        vc._geo = vc._geo.compute()
        self.assertFalse(dask.is_dask_collection(vc.data))
        self.assertTrue(dask.is_dask_collection(vc.to_lazy().data))

    def test_to_dataframe(self):
        self.assertIsInstance(self.v.to_dataframe(), dgpd.GeoDataFrame)
        self.assertIsInstance(
            self.v.to_lazy().to_dataframe(), dgpd.GeoDataFrame
        )
        self.assertTrue(
            self.v.data.compute().equals(self.v.to_dataframe().compute())
        )

    def test_to_crs(self):
        crs = rio.crs.CRS.from_epsg(4326)
        self.assertTrue(self.v.crs == rio.crs.CRS.from_epsg(5070))
        self.assertTrue(self.v.to_crs(crs).crs == crs)
        self.assertTrue(self.v.to_crs(4326).crs == crs)
        self.assertTrue(self.v.to_crs("epsg:4326").crs == crs)


class TestCastField(unittest.TestCase):
    def test_cast_field(self):
        v = open_vectors("tests/data/vector/pods.shp")
        dtypes = [int, "int16", np.int32, float, "float"]
        for d in dtypes:
            dt = np.dtype(d)
            vc = v.cast_field("POD_Num", d)
            self.assertTrue(vc.data.POD_Num.dtype == dt)


@pytest.mark.parametrize(
    "vec",
    [
        "tests/data/vector/pods.shp",
        open_vectors("tests/data/vector/pods.shp"),
        open_vectors("tests/data/vector/pods.shp").data.repartition(
            npartitions=5
        ),
        open_vectors("tests/data/vector/pods.shp")
        .data.repartition(npartitions=5)
        .spatial_shuffle(),
    ],
)
def test_add_objectid_column(vec):
    if isinstance(vec, str):
        vec = open_vectors(vec)
    geoms_df = vec.data if isinstance(vec, Vector) else vec

    id_values = []
    for i, part in enumerate(geoms_df.partitions):
        part = part.compute()
        base = i * _OBJECTID_BASE
        id_values.extend(np.arange(1, len(part) + 1) + base)
    id_values = np.array(id_values)

    result = rts.vector.add_objectid_column(vec, name="id_col")
    result_df = result
    if isinstance(vec, Vector):
        assert isinstance(result, Vector)
        result_df = result.data

    if geoms_df.spatial_partitions is not None:
        assert result.spatial_partitions.equals(geoms_df.spatial_partitions)
    assert geoms_df.npartitions == result_df.npartitions
    assert "id_col" in result_df
    assert np.allclose(id_values, result_df.id_col.compute().to_numpy())



import os
from pathlib import Path

import dask
import numpy as np
import rasterio as rio
import rioxarray as xrio
import xarray as xr
from affine import Affine
from dask.array.core import normalize_chunks as dask_chunks

from raster_tools.dtypes import F32, F64, I64, U8, is_bool, is_float, is_int
from raster_tools.masking import create_null_mask
from raster_tools.utils import (
    is_strictly_decreasing,
    is_strictly_increasing,
    to_chunk_dict,
    validate_path,
)


class RasterIOError(BaseException):
    pass


class RasterDataError(BaseException):
    pass


def _get_extension(path):
    return os.path.splitext(path)[-1].lower()


def _get_chunking_info_from_file(src_file):
    with rio.open(src_file) as src:
        tile_shape = (1, *src.block_shapes[0])
        shape = (src.count, *src.shape)
        dtype = np.dtype(src.dtypes[0])
        return tile_shape, shape, dtype


def _get_chunks(data=None, src_file=None):
    chunks = (1, "auto", "auto")
    if data is None:
        if src_file is None:
            return chunks
        tile_shape, shape, dtype = _get_chunking_info_from_file(src_file)
    else:
        shape = data.shape
        dtype = data.dtype
        tile_shape = None
        if dask.is_dask_collection(data):
            tile_shape = data.chunks
        elif src_file is not None:
            _, tile_shape, _ = _get_chunking_info_from_file(src_file)
    return dask_chunks(chunks, shape, dtype=dtype, previous_chunks=tile_shape)


def chunk(xrs, src_file=None):
    chunks = to_chunk_dict(
        _get_chunks(
            xrs.raster if isinstance(xrs, xr.Dataset) else xrs, src_file
        )
    )
    return xrs.chunk(chunks)


TIFF_EXTS = frozenset((".tif", ".tiff"))
NC_EXTS = frozenset((".cdf", ".nc", ".nc4"))
HDF_EXTS = frozenset((".hdf", ".h4", ".hdf4", ".he2", ".h5", ".hdf5", ".he5"))
GRIB_EXTS = frozenset((".grib", ".grib2", ".grb", ".grb2", ".gb", ".gb2"))
BATCH_EXTS = frozenset((".bch",))

# File extenstions that can't be read in yet
READ_NOT_IMPLEMENTED_EXTS = NC_EXTS | HDF_EXTS | GRIB_EXTS
# File extenstions that can't be written out yet
WRITE_NOT_IMPLEMENTED_EXTS = NC_EXTS | HDF_EXTS | GRIB_EXTS

IO_UNDERSTOOD_TYPES = (str, Path)


def is_batch_file(path):
    return _get_extension(path) in BATCH_EXTS


def normalize_xarray_data(xrs):
    if len(xrs.shape) > 3 or len(xrs.shape) < 2:
        raise ValueError(
            "Invalid shape. xarray.DataArray objects must have 2D or 3D "
            "shapes."
        )
    if len(xrs.shape) == 2:
        # Add band dim
        xrs = xrs.expand_dims({"band": [1]})
    dims = xrs.dims
    if "lon" in dims:
        xrs = xrs.rename({"lon": "x"})
        dims = xrs.dims
    if "lat" in dims:
        xrs = xrs.rename({"lat": "y"})
        dims = xrs.dims
    if dims != ("band", "y", "x"):
        # No easy way to figure out how best to transpose based on dim names so
        # just assume the order is valid and rename.
        xrs = xrs.rename(
            {
                d: new_d
                for d, new_d in zip(dims, ("band", "y", "x"))
                if d != new_d
            }
        )
    if xrs.band.to_numpy()[0] != 1:
        xrs["band"] = np.arange(1, len(xrs.band) + 1)
    if any(dim not in xrs.coords for dim in xrs.dims):
        raise ValueError(
            "Invalid coordinates on xarray.DataArray object:\n{xrs!r}"
        )
    # Make sure that x and y are always increasing. xarray will auto align
    # rasters but when a raster is converted to a numpy or dask array, the
    # data may not be aligned. This ensures that rasters converted to
    # non-georeferenecd formats will be oriented the same.
    if is_strictly_decreasing(xrs.x):
        xrs = xrs.isel(x=slice(None, None, -1))
    if is_strictly_increasing(xrs.y):
        xrs = xrs.isel(y=slice(None, None, -1))
    tf = xrs.rio.transform(True)
    xrs = xrs.rio.write_transform(tf)
    return xrs


ESRI_DEFAULT_F32_NV = np.finfo(F32).min


def normalize_null_value(nv, dtype):
    # Make sure that ESRI's default F32 null value is properly
    # registered as F32
    if dtype == F32 and nv is not None and np.isclose(nv, ESRI_DEFAULT_F32_NV):
        nv = F32.type(nv)
    # Some rasters have (u)int dtype and a null value that is a whole number
    # but it gets read in as a float. This can cause a lot of accidental type
    # promotions down the pipeline. Check for this case and correct it.
    if is_int(dtype) and is_float(nv) and float(nv).is_integer():
        nv = int(nv)
    return nv


def open_raster_from_path(path):
    if type(path) in IO_UNDERSTOOD_TYPES:
        path = str(path)
        path = os.path.abspath(path)
    else:
        raise RasterIOError(
            f"Could not resolve input to a raster path: '{path}'"
        )
    validate_path(path)
    ext = _get_extension(path)

    xrs = None
    # Try to let gdal open anything but NC, HDF, GRIB files
    if not ext or ext not in READ_NOT_IMPLEMENTED_EXTS:
        try:
            xrs = xrio.open_rasterio(path, chunks=to_chunk_dict(_get_chunks()))
        except rio.errors.RasterioIOError as e:
            raise RasterIOError(
                "Could not open given path as a raster."
            ) from e
    elif ext in READ_NOT_IMPLEMENTED_EXTS:
        raise NotImplementedError(
            "Reading of NetCDF, HDF, and GRIB files is not supported at this"
            " time."
        )
    else:
        raise RasterIOError("Unknown file type")
    if isinstance(xrs, xr.Dataset):
        raise RasterDataError("Too many data variables in input data")
    assert isinstance(
        xrs, xr.DataArray
    ), "Resulting data structure must be a DataArray"
    if not dask.is_dask_collection(xrs):
        xrs = chunk(xrs, path)

    xrs = normalize_xarray_data(xrs)

    nv = xrs.attrs.get("_FillValue", None)
    nv = normalize_null_value(nv, xrs.dtype)
    mask = create_null_mask(xrs, nv)
    return xrs, mask, nv


def write_raster(xrs, path, no_data_value, **rio_gdal_kwargs):
    ext = _get_extension(path)
    rio_is_bool = False
    if ext in TIFF_EXTS or len(ext) == 0:
        if xrs.dtype == I64:
            # GDAL, and thus rioxarray and rasterio, doesn't support I64 so
            # cast up to float. This avoids to_raster throwing a TypeError.
            xrs = xrs.astype(F64)
        elif is_bool(xrs.dtype):
            # GDAL doesn't support boolean dtype either so convert to uint8
            # 0-1 encoding.
            rio_is_bool = True
            xrs = xrs.astype(U8)

    if not ext or ext not in WRITE_NOT_IMPLEMENTED_EXTS:
        kwargs = {"lock": True, "compute": True, **rio_gdal_kwargs}
        if "blockheight" in kwargs:
            value = kwargs.pop("blockheight")
            kwargs["blockysize"] = value
        if "blockwidth" in kwargs:
            value = kwargs.pop("blockwidth")
            kwargs["blockxsize"] = value
        if rio_is_bool:
            # Store each entry using a single bit
            kwargs["nbits"] = 1
        xrs.rio.to_raster(path, **kwargs)
    else:
        # TODO: populate
        raise NotImplementedError()


class AffineEncodingError(Exception):
    pass


class DimensionsError(Exception):
    pass


def _get_valid_variables(meta, ignore_too_many_dims):
    data_vars = list(meta.data_vars)
    valid = []
    for v in data_vars:
        n = meta[v].squeeze().ndim
        if n > 3:
            if ignore_too_many_dims:
                continue
            else:
                raise DimensionsError(
                    f"Too many dimensions for variable {v!r} with "
                    f"{meta[v].ndim}."
                )
        elif n in (2, 3):
            valid.append(v)
        else:
            raise DimensionsError(
                f"Too few dimensions for variable {v!r} with {n}."
            )
    if not valid:
        raise ValueError("No valid raster variables found")
    return valid


def _build_xr_raster(path, variable, affine, crs, xarray_kwargs):
    if affine is None:
        affine = Affine(1, 0, 0, 0, -1, 0, 0)
    kwargs = xarray_kwargs.copy()
    kwargs["chunks"] = "auto"
    var = xr.open_dataset(path, **kwargs)[variable].squeeze()
    var_data = var.data
    if var.ndim == 2:
        var_data = np.expand_dims(var_data, axis=0)
    var_data = var_data.rechunk((1, "auto", "auto"))
    band = np.array(list(range(var_data.shape[0])))
    x = var[var.rio.x_dim].to_numpy()
    y = var[var.rio.y_dim].to_numpy()
    new_var = xr.DataArray(
        var_data, dims=["band", "y", "x"], coords=(band, y, x)
    )
    new_var = new_var.rio.write_transform(affine)
    if crs is not None:
        new_var = new_var.rio.write_crs(crs)
    nv = var._FillValue if "_FillValue" in var.attrs else var.rio.nodata
    if nv is not None:
        new_var = new_var.rio.write_nodata(nv)
    return new_var


def _get_affine(ds):
    try:
        affine = ds.rio.transform()
    except TypeError as err:
        # Some datasets like gridMET improperly encode the transform.
        raise AffineEncodingError(
            "Error reading GeoTransform data:"
            f"{ds.coords[ds.rio.grid_mapping].attrs['GeoTransform']!r}"
        ) from err
    return affine


def open_dataset(
    path,
    crs=None,
    ignore_extra_dim_errors=False,
    xarray_kwargs=None,
):
    """Open a netCDF or GRIB dataset.

    This function opens a netCDF or GRIB dataset file and returns a dictionary
    of Raster objectds where each raster corrersponds to the variables in the
    the file. netCDF/GRIB files can be N-dimensional, while rasters only
    comprehend 2 to 3 dimensions (band, y, x), so it may not be possible to map
    all variables in a file to a raster. See the `ignore_extra_dim_errors`
    option below for more information.

    Parameters
    ----------
    path : str
        THe path to the netCDF or GRIB dataset file.
    crs : str, rasterio.crs.CRS, optional
        A coordinate reference system definition to attach to the dataset. This
        can be an EPSG, PROJ, or WKT string. It can also be a
        `rasterio.crs.CRS` object. netCDF/GRIB files do not always encode a
        CRS. This option allows a CRS to be supplied, if known ahead of time.
        It can also be used to override the CRS encoded in the file.
    ignore_extra_dim_errors : bool, optional
        If ``True``, ignore dataset variables that cannot be mapped to a
        raster. An error is raised, otherwise. netCDF/GRIB files allow
        N-dimensional. Rasters only comprehend 2 or 3 dimensional data so it is
        not always possible to map a variable to a raster. The default is
        ``False``.
    xarray_kwargs : dict, optional
        Keyword arguments to supply to `xarray.open_dataset` when opening the
        file.

    Raises
    ------
    raster_tools.io.AffineEncodingError
        Raised if the affine matrix is improperly encoded.
    ra

    Returns
    -------
    dataset : dict of Raster
        A ``dict`` of Raster objects. The keys are the variable names in the
        dataset file and the values are the corresponding variable data as a
        raster.

    """
    from raster_tools.raster import Raster

    if xarray_kwargs is None:
        xarray_kwargs = {}
    xarray_kwargs["decode_coords"] = "all"
    tmp_ds = xr.open_dataset(path, **xarray_kwargs)
    data_vars = _get_valid_variables(tmp_ds, ignore_extra_dim_errors)
    crs = crs or tmp_ds.rio.crs
    affine = _get_affine(tmp_ds)
    tmp_ds = None
    ds = {}
    for v in data_vars:
        var = _build_xr_raster(path, v, affine, crs, xarray_kwargs)
        ds[v] = Raster(var)
    return ds



# isort: off
# TODO(pygeos): remove this once shapely is the default backend for geopandas.
# Force raster_tools._compat to be loaded before geopandas when running tests
import raster_tools as rts  # noqa: F401

# isort: on

import dask
import dask.dataframe as dd
import dask_geopandas as dgpd
import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import scipy
import shapely

from raster_tools import band_concat
from raster_tools.zonal import (
    _asm_agg,
    _entropy_agg,
    _mode_agg,
    _nunique_agg,
    _raster_to_series,
    extract_points_eager,
    zonal_stats,
)
from tests import testdata
from tests.utils import arange_raster


def asm(x):
    return rts.stat_common.nanasm_jit(x.to_numpy())


def entropy(x):
    return rts.stat_common.nanentropy_jit(x.to_numpy())


def mode(x):
    m = scipy.stats.mode(x.to_numpy()).mode
    if np.isscalar(m):
        return m
    return m[0]


def nunique(x):
    return rts.stat_common.nan_unique_count_jit(x.to_numpy())


group_df1 = pd.DataFrame(
    {"zone": np.arange(5).repeat(4), "band_1": np.arange(20).astype("float64")}
)
group_df2 = pd.DataFrame(
    {
        "zone": [8, 0, 1, 1, 2, 3, 3, 3, 3, 3, 4, 4, 6, 7, 7, 7, 7, 7, 8, 8],
        "band_1": np.arange(20).astype("float64"),
    }
)
group_df3 = pd.DataFrame(
    {"zone": np.arange(5).repeat(4), "band_1": np.ones(20).astype("float64")}
)


@pytest.mark.parametrize(
    "stat,stat_truth",
    [
        (_asm_agg, asm),
        (_entropy_agg, entropy),
        (_mode_agg, mode),
        (_nunique_agg, nunique),
    ],
)
@pytest.mark.parametrize(
    "frame",
    [
        dd.from_pandas(group_df1, npartitions=1),
        dd.from_pandas(group_df1, npartitions=3),
        dd.from_pandas(group_df1, npartitions=10),
        dd.from_pandas(group_df2, npartitions=3),
        dd.from_pandas(group_df2, npartitions=7),
        dd.from_pandas(group_df3, npartitions=7),
    ],
)
def test_custom_stats(stat, stat_truth, frame):
    gdfc = frame.compute().groupby("zone")
    gdf = frame.groupby("zone")
    truth = gdfc.agg([stat_truth]).sort_index()
    result = gdf.agg([stat]).compute().sort_index()

    assert truth.equals(result)


def rasters_to_zonal_df(feat_raster, data_raster):
    feat_raster = feat_raster.chunk(data_raster.data.chunks)
    dfs = [_raster_to_series(feat_raster).rename("zone").compute()]
    for b in range(1, data_raster.nbands + 1):
        nv = data_raster.null_value
        band = (
            _raster_to_series(data_raster.get_bands(b))
            .rename(f"band_{b}")
            .compute()
        )
        if nv is not None:
            band = band.replace(data_raster.null_value, np.nan)
        # Cast up to avoid floating point issues in sums
        band = band.astype("float64")
        dfs.append(band)
    return pd.concat(dfs, axis=1)


@pytest.mark.parametrize(
    "stats",
    [
        ["max", "mean", "median", "min", "size", "std", "sum", "var"],
        ["asm", "entropy", "mode", "nunique", "mean", "var", "count"],
        ["asm", "median"],
        "median",
        ["mean", "median"],
    ],
)
@pytest.mark.parametrize(
    "features,raster",
    [
        # No null data in data raster
        (
            testdata.vector.pods_small.data.repartition(npartitions=3),
            testdata.raster.dem_small.chunk((1, 20, 20)),
        ),
        # Some null data in data raster outside of features
        (
            testdata.vector.pods_small,
            rts.clipping.clip(
                testdata.vector.pods_small.buffer(100),
                testdata.raster.dem_small.chunk((1, 20, 20)),
            ),
        ),
        # All null data in data raster outside of features and some inside
        (
            testdata.vector.pods_small,
            rts.clipping.clip(
                testdata.vector.pods_small.data.compute()
                .dissolve()
                .buffer(-100),
                testdata.raster.dem_small.chunk((1, 20, 20)),
            ),
        ),
        # Raster features instead of vector
        (
            testdata.vector.pods_small.to_raster(testdata.raster.dem_small),
            testdata.raster.dem_small.chunk((1, 20, 20)),
        ),
        # Raster features instead of vector and mismatched chunksize
        (
            testdata.vector.pods_small.to_raster(
                testdata.raster.dem_small.chunk((1, 25, 25))
            ),
            testdata.raster.dem_small.chunk((1, 20, 20)),
        ),
        # No null value set
        (
            testdata.vector.pods_small,
            testdata.raster.dem_small.set_null_value(None).chunk((1, 20, 20)),
        ),
    ],
)
def test_zonal_stats(features, raster, stats):
    stats_working = [stats] if isinstance(stats, str) else stats.copy()
    for i, s in enumerate(stats_working):
        if s == "asm":
            stats_working[i] = asm
        elif s == "entropy":
            stats_working[i] = entropy
        elif s == "mode":
            stats_working[i] = mode
        elif s == "nunique":
            stats_working[i] = nunique

    if not isinstance(features, rts.Raster):
        feat_raster = rts.rasterize.rasterize(features, raster)
    else:
        feat_raster = features
    tdf = rasters_to_zonal_df(feat_raster, raster)
    # Trim off areas outside features
    tdf = tdf[tdf.zone != feat_raster.null_value]
    truth_df = tdf.groupby("zone").agg(stats_working).sort_index()

    result = zonal_stats(features, raster, stats)
    assert isinstance(result, dd.DataFrame)
    assert result.columns.equals(truth_df.columns)
    assert result.dtypes.equals(truth_df.dtypes)
    resultc = result.compute().sort_index()
    # Use this because .equals() is too sensitive
    assert (truth_df - resultc).abs().max().max() < 1e-8


@pytest.mark.parametrize(
    "raster",
    [
        testdata.raster.dem_small,
        rts.band_concat(
            [
                testdata.raster.dem_small,
                testdata.raster.dem_small,
                testdata.raster.dem_small,
            ]
        ).chunk((1, 20, 20)),
    ],
)
def test_zonal_stats_long_format(raster):
    zdf = zonal_stats(
        testdata.vector.pods_small, raster, ["mean", "min", "max"]
    )
    truth = (
        zdf.compute()  # noqa: PD013  warning is wrong. .melt cannot do this
        .stack(0, future_stack=True)
        .reset_index()
        .rename(columns={"level_1": "band"})
        .sort_values(["band", "zone"])
        .sort_index()
    )
    lookup = {f"band_{i + 1}": i + 1 for i in range(raster.nbands)}
    truth["band"] = truth.band.apply(lambda x: lookup[x])
    zdf = zonal_stats(
        testdata.vector.pods_small,
        raster,
        ["mean", "min", "max"],
        wide_format=False,
    )

    assert truth.columns.equals(zdf.columns)
    zdf = zdf.compute().sort_index()
    assert truth.equals(zdf)


def get_random_points(n, nparts, dem):
    xmin, ymin, xmax, ymax = dem.bounds
    xspan = xmax - xmin
    yspan = ymax - ymin
    x = xmin + (xspan * 1.1 * np.random.random(n)) - (0.05 * xspan)
    y = ymin + (yspan * 1.1 * np.random.random(n)) - (0.05 * yspan)
    points = gpd.GeoSeries.from_xy(x, y, crs=dem.crs).to_frame("geometry")
    return dgpd.from_geopandas(points, npartitions=nparts)


def dem_clipped_small():
    return testdata.raster.dem_clipped_small


@pytest.mark.parametrize(
    "dem,n,nparts,name",
    [
        (dem_clipped_small(), 1_000, 1, None),
        (dem_clipped_small(), 1_000, 4, "values"),
        (dem_clipped_small().chunk((1, 10, 10)), 1_000, 4, "points"),
        (
            band_concat(
                [
                    dem_clipped_small(),
                    dem_clipped_small() + 1,
                    dem_clipped_small() + 10,
                ]
            ).chunk((1, 10, 10)),
            1_000,
            4,
            None,
        ),
        (arange_raster((700, 6, 6)), 10, 1, None),
        (arange_raster((700, 6, 6)).set_crs("EPSG:3857"), 10, 1, None),
    ],
)
def test_extract_points_eager(dem, n, nparts, name):
    if name is None:
        name = "extracted"
    points_df = get_random_points(n, nparts, dem)
    points = points_df.geometry.compute().to_list()
    x, y = dask.compute(points_df.geometry.x, points_df.geometry.y)
    r, c = dem.index(x, y)
    data = dem.to_numpy()
    mask = dem.mask.compute()
    nb, nr, nc = data.shape
    valid = (r >= 0) & (r < nr) & (c >= 0) & (c < nc)
    n = len(x)
    bbox = shapely.geometry.box(*dem.bounds)
    point_check = [bbox.contains(p) for p in points]
    assert np.allclose(valid, point_check)
    dfs = []
    for bnd in range(nb):
        d = {
            "band": np.full(n, bnd + 1, dtype=np.min_scalar_type(bnd + 1)),
            name: np.full(n, np.nan),
        }
        extracted = data[bnd, r[valid], c[valid]].astype(float)
        masked = mask[bnd, r[valid], c[valid]]
        extracted[masked] = np.nan
        d[name][valid] = extracted
        dfs.append(
            pd.DataFrame(d, index=pd.RangeIndex(n * bnd, n * (bnd + 1)))
        )
    truth = pd.concat(dfs)
    result = extract_points_eager(points_df, dem, name)
    resultc = result.compute()
    assert isinstance(result, dask.dataframe.DataFrame)
    assert isinstance(resultc, pd.DataFrame)
    assert result.band.dtype == np.min_scalar_type(nb)
    assert result.known_divisions
    assert result.npartitions == nb
    assert all(result.columns == ["band", name])
    assert truth.equals(resultc)



import numpy as np
import pytest
import xarray as xr

from raster_tools.io import DimensionsError, open_dataset
from raster_tools.masking import get_default_null_value
from tests.utils import assert_valid_raster


@pytest.mark.parametrize(
    "path,crs",
    [
        ("tests/data/raster/grib_example_1.grib", None),
        ("tests/data/raster/grib_example_2.grib2", None),
        ("tests/data/raster/nc_example_1.nc", None),
        ("tests/data/raster/nc_example_2.nc", None),
        ("tests/data/raster/nc_example_3.nc", "EPSG:4326"),
    ],
)
def test_open_dataset(path, crs):
    # mask_and_scale is on by default for netcdf files now. This replaces all
    # _FillValue occurances with nan and drops the _FIllValue attribute. Set it
    # to false so that we can make sure the fill value is being detected.
    meta = xr.open_dataset(path, decode_coords="all", mask_and_scale=False)
    vars_ = list(meta.data_vars)
    vars_ = [v for v in vars_ if meta[v].squeeze().ndim <= 3]

    result = open_dataset(
        path,
        ignore_extra_dim_errors=True,
        # See comment above about turning off mask_and_scale
        xarray_kwargs={"mask_and_scale": False},
    )

    assert isinstance(result, dict)
    assert len(result) == len(vars_)
    assert sorted(vars_) == sorted(result)
    for k, v in result.items():
        assert_valid_raster(v)
        assert v.crs == crs
        assert v.affine is not None
        fill = meta[k].rio.nodata
        if fill is None:
            fill = np.nan
        if not np.isnan(fill):
            assert v.null_value == fill
        else:
            assert v.null_value == get_default_null_value(v.dtype)
        assert v.xdata.dims == ("band", "y", "x")
        truth_data = meta[k].squeeze().compute().to_numpy()
        ycoord = meta[meta.rio.y_dim].to_numpy()
        if ycoord[0] < ycoord[-1]:
            truth_data = truth_data[..., ::-1, :]
        data = v.data.compute().squeeze()
        assert np.allclose(truth_data, data)


def test_open_dataset_extra_dims_error():
    with pytest.raises(DimensionsError):
        open_dataset("tests/data/raster/nc_example_2.nc")
    assert (
        open_dataset(
            "tests/data/raster/nc_example_2.nc", ignore_extra_dim_errors=True
        )
        is not None
    )


def test_open_dataset_crs():
    assert xr.open_dataset("tests/data/raster/nc_example_1.nc").rio.crs is None
    dataset = open_dataset(
        "tests/data/raster/nc_example_1.nc", crs="EPSG:4326"
    )
    assert dataset["t2m"].crs == "EPSG:4326"



"""Module for providing test data."""
import glob
import os

import raster_tools as rts


class _DataProvider:
    def __init__(self, opener, data_dir, data_aliases, preferred_ext=None):
        self._opener = opener
        self._data_dir = data_dir
        self._data_aliases = data_aliases
        self._preferred_ext = preferred_ext

    def __getattr__(self, key):
        if key in self._data_aliases:
            value = self._data_aliases[key]
            if callable(value):
                return value(self._opener)
            return self._opener(value)

        # Try to find matching file
        paths = glob.glob(os.path.join(self._data_dir, f"{key}.*"))
        if len(paths):
            if self._preferred_ext is not None:
                for p in paths:
                    if os.path.splitext(p)[1] == self._preferred_ext:
                        return self._opener(p)
            # Try grabbing first one.
            return self._opener(paths[0])
        else:
            raise ValueError(f"Could not provide data for {key!r}")

    def open_file(self, file, *args, **kwargs):
        return self._opener(
            os.path.join(self._data_dir, file), *args, **kwargs
        )


DATA_ROOT = os.path.join(os.path.dirname(__file__), "data")
RASTER_DIR = os.path.join(DATA_ROOT, "raster")
VECTOR_DIR = os.path.join(DATA_ROOT, "vector")


_raster_aliases = {
    "dem_aspect": os.path.join(RASTER_DIR, "aspect.tif"),
    "dem_curv": os.path.join(RASTER_DIR, "curv.tif"),
    "dem_easting": os.path.join(RASTER_DIR, "easting.tif"),
    "dem_hillshade": os.path.join(RASTER_DIR, "hillshade.tif"),
    "dem_northing": os.path.join(RASTER_DIR, "northing.tif"),
    "dem_slope": os.path.join(RASTER_DIR, "slope.tif"),
    "dem_slope_percent": os.path.join(RASTER_DIR, "slope_percent.tif"),
}
raster = _DataProvider(rts.Raster, RASTER_DIR, _raster_aliases, ".tif")

_vector_aliases = {
    "pods": os.path.join(VECTOR_DIR, "pods.shp"),
    "lmus": lambda x: x(os.path.join(VECTOR_DIR, "Zones.gdb"), leyers="LMUs"),
}
vector = _DataProvider(rts.open_vectors, VECTOR_DIR, _vector_aliases, ".shp")



import numpy as np
import pytest
from odc.geo.geobox import GeoBox

from raster_tools import warp
from raster_tools.masking import get_default_null_value
from tests import testdata
from tests.utils import assert_valid_raster


def _repoject(raster, crs, method):
    nv = (
        raster.null_value
        if raster._masked
        else get_default_null_value(raster.dtype)
    )
    xreprojected = raster.xdata.odc.reproject(
        crs, resampling=method, dst_nodata=nv
    )
    xmask = xreprojected == nv
    return xreprojected.rio.write_nodata(nv), xmask


@pytest.mark.parametrize(
    "method",
    [
        "nearest",
        "bilinear",
        "cubic",
        "cubic_spline",
        "lanczos",
        "average",
        "mode",
        "max",
        "min",
        "med",
        "q1",
        "q3",
        "sum",
        "rms",
    ],
)
@pytest.mark.parametrize(
    "raster",
    [
        testdata.raster.dem_small.chunk((1, 20, 20)),
        testdata.raster.dem_small.set_null_value(None),
        testdata.raster.dem_small.chunk((1, 20, 20))
        .remap_range((0, 1100, -1))
        .set_null_value(-1),
    ],
)
def test_reproject(raster, method):
    crs = "EPSG:5070"
    truth_reprojected, truth_mask = _repoject(raster, crs, method)

    result = warp.reproject(raster, crs, method)
    assert_valid_raster(result)
    assert result.crs == crs
    assert result.null_value == truth_reprojected.rio.nodata
    assert result.data.chunksize == raster.data.chunksize
    assert np.allclose(result.xdata, truth_reprojected)
    assert np.allclose(result.xmask, truth_mask)

    result = raster.reproject(crs, method)
    assert_valid_raster(result)
    assert result.crs == crs
    assert result.null_value == truth_reprojected.rio.nodata
    assert result.data.chunksize == raster.data.chunksize
    assert np.allclose(result.xdata, truth_reprojected)
    assert np.allclose(result.xmask, truth_mask)


@pytest.mark.parametrize(
    "crs_or_geobox",
    [
        5070,
        "EPSG:5070",
        testdata.raster.dem_small.geobox.to_crs(4326),
        testdata.raster.dem_small.geobox.zoom_to(resolution=15),
    ],
)
def test_reproject_crs(crs_or_geobox):
    raster = testdata.raster.dem_small
    if isinstance(crs_or_geobox, GeoBox):
        dst_geobox = crs_or_geobox
    else:
        dst_geobox = raster.geobox.to_crs(crs_or_geobox)
    truth_reprojected, truth_mask = _repoject(
        raster, crs_or_geobox, "bilinear"
    )
    crs = (
        crs_or_geobox.crs
        if isinstance(crs_or_geobox, GeoBox)
        else crs_or_geobox
    )

    result = warp.reproject(raster, crs_or_geobox, resample_method="bilinear")
    assert_valid_raster(result)
    if crs == 4326:
        # pixel -> world transforms are limited in their precision at fine
        # scales like 1m. The result of a reprojection from some CRS to
        # lat/lon will cause tiny differences in the resulting affine matrix.
        # see: https://github.com/opendatacube/odc-geo/issues/127
        assert result.geobox.crs == crs
        assert result.geobox.shape == dst_geobox.shape
        assert np.allclose(list(result.geobox.affine), list(dst_geobox.affine))
        # rioxarray pulls the correct affine matrix from the grid_mapping
        # attribute
        assert result.xdata.rio.transform() == dst_geobox.affine
    else:
        assert result.geobox == dst_geobox
    assert result.null_value == truth_reprojected.rio.nodata
    assert np.allclose(result.to_numpy(), truth_reprojected.to_numpy())
    assert np.allclose(result.xmask, truth_mask)


@pytest.mark.parametrize(
    "resolution",
    [10, 30, 50, 55.5],
)
def test_reproject_resolution(resolution):
    raster = testdata.raster.dem_small
    gb = raster.geobox.zoom_to(resolution=resolution)
    truth_reprojected, truth_mask = _repoject(raster, gb, "bilinear")

    result = warp.reproject(
        raster, resolution=resolution, resample_method="bilinear"
    )
    assert_valid_raster(result)
    assert result.crs == raster.crs
    assert result.resolution == gb.resolution.xy
    assert result.null_value == truth_reprojected.rio.nodata
    assert np.allclose(result.to_numpy(), truth_reprojected.to_numpy())
    assert np.allclose(result.xmask, truth_mask)

    result = raster.reproject(
        resolution=resolution, resample_method="bilinear"
    )
    assert_valid_raster(result)
    assert result.crs == raster.crs
    assert result.resolution == gb.resolution.xy
    assert result.null_value == truth_reprojected.rio.nodata
    assert np.allclose(result.to_numpy(), truth_reprojected.to_numpy())
    assert np.allclose(result.xmask, truth_mask)


@pytest.mark.parametrize(
    "crs,resolution",
    [
        (5070, None),
        (None, 15),
        (5070, 15),
        (testdata.raster.dem_small.geobox.to_crs(5070), 15),
    ],
)
def test_reproject_crs_and_resolution(crs, resolution):
    raster = testdata.raster.dem_small

    if crs is None:
        target_grid = raster.geobox
    elif isinstance(crs, GeoBox):
        target_grid = crs
    else:
        target_grid = raster.geobox.to_crs(crs)
    if resolution is not None:
        target_grid = target_grid.zoom_to(resolution=resolution)

    truth_reprojected, truth_mask = _repoject(raster, target_grid, "bilinear")

    result = warp.reproject(
        raster, crs, resolution=resolution, resample_method="bilinear"
    )
    assert_valid_raster(result)
    assert result.crs == target_grid.crs
    assert result.resolution == target_grid.resolution.xy
    assert result.null_value == truth_reprojected.rio.nodata
    assert np.allclose(result.to_numpy(), truth_reprojected.to_numpy())
    assert np.allclose(result.xmask, truth_mask)

    result = raster.reproject(
        crs, resolution=resolution, resample_method="bilinear"
    )
    assert_valid_raster(result)
    assert result.crs == target_grid.crs
    assert result.resolution == target_grid.resolution.xy
    assert result.null_value == truth_reprojected.rio.nodata
    assert np.allclose(result.to_numpy(), truth_reprojected.to_numpy())
    assert np.allclose(result.xmask, truth_mask)



import dask.array as da
import numpy as np
import pytest

from raster_tools import creation
from raster_tools.masking import get_default_null_value
from tests.utils import (
    arange_raster,
    assert_rasters_similar,
    assert_valid_raster,
)


def templates():
    ts = []
    r = arange_raster((2, 10, 10)).set_null_value(-1).set_crs("EPSG:5070")
    r._ds.mask.data[0, :3, :3] = True
    r._ds.mask.data[1, -3:, -3:] = True
    r = r.chunk((1, 3, 3))
    ts.append(r)
    ts.append(arange_raster((3, 100, 100), dtype=float))
    return ts


@pytest.mark.parametrize("copy_mask", [0, 1])
@pytest.mark.parametrize("nbands", [1, 2, 3])
@pytest.mark.parametrize(
    "dist,params",
    [
        ("binomial", (5, 0.5)),
        ("b", (5, 0.5)),
        ("normal", (0, 20)),
        ("n", (1, 1)),
        ("uniform", ()),
        ("u", ()),
        # Make sure params are ignored
        ("u", (1, 2)),
        ("weibull", (11,)),
        ("w", (11,)),
    ],
)
@pytest.mark.parametrize("template", templates())
def test_random_raster(template, dist, params, nbands, copy_mask):
    if nbands == template.nbands:
        mask_truth = template.mask.compute()
    else:
        mask_truth = da.stack([template.mask[0]] * nbands).compute()
    assert mask_truth.ndim == 3
    assert mask_truth.shape[0] == nbands

    result = creation.random_raster(
        template,
        distribution=dist,
        bands=nbands,
        params=params,
        copy_mask=copy_mask,
    )
    assert_valid_raster(result)
    assert_rasters_similar(template, result, check_nbands=False)
    assert result.nbands == nbands
    result.eval()
    if copy_mask and template._masked:
        assert result._masked
        assert result.null_value == get_default_null_value(result.dtype)
        assert np.allclose(result.mask.compute(), mask_truth)
        assert (result.data[result.mask].compute() == result.null_value).all()
    else:
        assert not result._masked
        assert result.mask.sum().compute() == 0


def run_constant_raster_tests(
    op_to_test, template, value, nbands, dtype, copy_mask, pass_value
):
    dtype = np.dtype(dtype) if dtype is not None else template.dtype
    if nbands == template.nbands:
        mask_truth = template.mask.compute()
    else:
        mask_truth = da.stack([template.mask[0]] * nbands).compute()
    assert mask_truth.ndim == 3
    assert mask_truth.shape[0] == nbands

    kwargs = {"bands": nbands, "dtype": dtype, "copy_mask": copy_mask}
    if pass_value:
        kwargs["value"] = value
    result = op_to_test(template, **kwargs)
    assert_valid_raster(result)
    assert_rasters_similar(template, result, check_nbands=False)
    assert result.nbands == nbands
    assert result.dtype == dtype
    result.eval()
    if copy_mask and template._masked:
        assert result._masked
        if dtype == template.dtype:
            assert result.null_value == template.null_value
        else:
            assert result.null_value == get_default_null_value(result.dtype)
        assert np.allclose(result.mask.compute(), mask_truth)
        assert (result.data[result.mask].compute() == result.null_value).all()
    else:
        assert not result._masked
        assert result.mask.sum().compute() == 0
    if value is not None:
        assert (result == value).all().compute()


@pytest.mark.parametrize("copy_mask", [0, 1])
@pytest.mark.parametrize("dtype", [None, "int32"])
@pytest.mark.parametrize("nbands", [1, 2, 3])
@pytest.mark.parametrize("template", templates())
def test_empty_like(template, nbands, dtype, copy_mask):
    run_constant_raster_tests(
        creation.empty_like, template, None, nbands, dtype, copy_mask, False
    )


@pytest.mark.parametrize("copy_mask", [0, 1])
@pytest.mark.parametrize("dtype", [None, "int32"])
@pytest.mark.parametrize("nbands", [1, 2, 3])
@pytest.mark.parametrize("value", [9, 100, -10])
@pytest.mark.parametrize("template", templates())
def test_full_like(template, value, nbands, dtype, copy_mask):
    run_constant_raster_tests(
        creation.full_like, template, value, nbands, dtype, copy_mask, True
    )


@pytest.mark.parametrize("copy_mask", [0, 1])
@pytest.mark.parametrize("dtype", [None, "int32"])
@pytest.mark.parametrize("nbands", [1, 2, 3])
@pytest.mark.parametrize("template", templates())
def test_zeros_like(template, nbands, dtype, copy_mask):
    run_constant_raster_tests(
        creation.zeros_like, template, 0, nbands, dtype, copy_mask, False
    )


@pytest.mark.parametrize("copy_mask", [0, 1])
@pytest.mark.parametrize("dtype", [None, "int32"])
@pytest.mark.parametrize("nbands", [1, 2, 3])
@pytest.mark.parametrize("template", templates())
def test_ones_like(template, nbands, dtype, copy_mask):
    run_constant_raster_tests(
        creation.ones_like, template, 1, nbands, dtype, copy_mask, False
    )
